

/*
 * Class:        LinearRegression
 * Description:  General linear regression with the least squares method
 * Environment:  Java
 * Software:     SSJ
 * Copyright (C) 2013  Pierre L'Ecuyer and Université de Montréal
 * Organization: DIRO, Université de Montréal
 * @author       Richard Simard
 * @since        April 2013

 * SSJ is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License (GPL) as published by the
 * Free Software Foundation, either version 3 of the License, or
 * any later version.

 * SSJ is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * A copy of the GNU General Public License is available at
   <a href="http://www.gnu.org/licenses">GPL licence site</a>.
 */

package regres;

import cern.colt.matrix.DoubleMatrix2D;
import cern.colt.matrix.impl.DenseDoubleMatrix2D;
import cern.colt.matrix.linalg.QRDecomposition;


/**
 * This class implements different linear regression models, using the
 * least squares method to estimate the regression coefficients. Given
 * input data <SPAN CLASS="MATH"><I>x</I><SUB>i, j</SUB></SPAN> and response <SPAN CLASS="MATH"><I>y</I><SUB>i</SUB></SPAN>, we want to find coefficients <SPAN CLASS="MATH"><I>&#946;</I><SUB>j</SUB></SPAN>
 * that minimize the residuals of the form
 *
 * <P></P>
 * <DIV ALIGN="CENTER" CLASS="mathdisplay">
 * &sum;<SUB>i</SUB>(<I>y</I><SUB>i</SUB> - <I>&#946;</I><SUB>0</SUB> - &sum;<SUB>j=1</SUB><SUP>k</SUP><I>&#946;</I><SUB>j</SUB><I>x</I><SUB>i, j</SUB>)<SUP>2</SUP>.
 * </DIV><P></P>
 * Using matrix notation, one may write this equation as
 *
 * <P></P>
 * <DIV ALIGN="CENTER" CLASS="mathdisplay">
 * min<SUB><I>&#946;</I></SUB>| <I>Y</I> - <I>X&#946;</I>|<SUB>2</SUB>,
 * </DIV><P></P>
 * where the first column of <SPAN CLASS="MATH"><I>X</I></SPAN> is all 1, and the <SPAN CLASS="MATH"><I>L</I><SUB>2</SUB></SPAN> norm is used.
 *
 * <P>
 * Sometimes, one wants to use a basis of general functions <SPAN CLASS="MATH"><I>&#968;</I><SUB>j</SUB>(<I>t</I>)</SPAN>
 * with a minimization of the form
 *
 * <P></P>
 * <DIV ALIGN="CENTER" CLASS="mathdisplay">
 * &sum;<SUB>i</SUB>(<I>y</I><SUB>i</SUB> - &sum;<SUB>j=1</SUB><SUP>k</SUP><I>&#946;</I><SUB>j</SUB><I>&#968;</I><SUB>j</SUB>(<I>t</I><SUB>i</SUB>))<SUP>2</SUP>.
 * </DIV><P></P>
 * For example, we could have
 * <SPAN CLASS="MATH"><I>&#968;</I><SUB>j</SUB>(<I>t</I>) = <I>e</I><SUP>-<I>&#955;</I><SUB>j</SUB>t</SUP></SPAN> or some other
 * functions.
 * In that case, one has to choose the points <SPAN CLASS="MATH"><I>t</I><SUB>i</SUB></SPAN> at which to
 * compute the basis functions, and use a method below with
 *
 * <SPAN CLASS="MATH"><I>x</I><SUB>i, j</SUB> = <I>&#968;</I><SUB>j</SUB>(<I>t</I><SUB>i</SUB>)</SPAN>.
 *
 */
public class LinearRegression {



   /**
    * Computes the regression coefficients using the
    * least squares method. This is a simple linear regression with
    *  only 2 regression coefficients. The model is
    *
    * <P></P>
    * <DIV ALIGN="CENTER" CLASS="mathdisplay">
    * <I>y</I> = <I>&#945;</I> + <I>&#946;x</I>.
    * </DIV><P></P>
    * Given the <SPAN CLASS="MATH"><I>n</I></SPAN> data points
    * <SPAN CLASS="MATH">(<I>X</I><SUB>i</SUB>, <I>Y</I><SUB>i</SUB>)</SPAN>,
    * <SPAN CLASS="MATH"><I>i</I> = 0, 1,&#8230;,(<I>n</I> - 1)</SPAN>,
    *  the method computes and returns the array
    *
    * <SPAN CLASS="MATH">[<I>&#945;</I>, <I>&#946;</I>]</SPAN>.
    *
    * @param X the regressor variables
    *
    *    @param Y the response
    *
    *    @return the regression coefficients
    *
    */
   public static double[] calcCoefficients (double[] X, double[] Y) {
      if (X.length != Y.length)
         throw new IllegalArgumentException ("Lengths of X and Y are not equal");
      final int n = X.length;
      double[][] Xa = new double[n][1];
      for (int i = 0; i < n; i++)
         Xa[i][0] = X[i];

      return calcCoefficients0 (Xa, Y);
   }


   /**
    * Computes the regression coefficients using the
    * least squares method. This is a model for multiple linear regression.
    * There are <SPAN CLASS="MATH"><I>k</I> + 1</SPAN> regression coefficients <SPAN CLASS="MATH"><I>&#946;</I><SUB>j</SUB></SPAN>,
    * <SPAN CLASS="MATH"><I>j</I> = 0, 1,&#8230;, <I>k</I></SPAN> and
    * <SPAN CLASS="MATH"><I>k</I></SPAN> regressors variables <SPAN CLASS="MATH"><I>x</I><SUB>j</SUB></SPAN>. The model is
    *
    * <P></P>
    * <DIV ALIGN="CENTER" CLASS="mathdisplay">
    * <I>y</I> = <I>&#946;</I><SUB>0</SUB> + &sum;<SUB>j=1</SUB><SUP>k</SUP><I>&#946;</I><SUB>j</SUB><I>x</I><SUB>j</SUB>.
    * </DIV><P></P>
    * There are <SPAN CLASS="MATH"><I>n</I></SPAN> data points <SPAN CLASS="MATH"><I>Y</I><SUB>i</SUB></SPAN>, <SPAN CLASS="MATH"><I>X</I><SUB>i, j</SUB></SPAN>,
    * <SPAN CLASS="MATH"><I>i</I> = 0, 1,&#8230;,(<I>n</I> - 1)</SPAN>, and each
    * <SPAN CLASS="MATH"><I>X</I><SUB>i</SUB></SPAN> is a <SPAN CLASS="MATH"><I>k</I></SPAN>-dimensional point.
    * Given the response <TT>Y[i]</TT> and the regressor variables <TT>X[i][j]</TT>,
    *
    * <SPAN CLASS="MATH"><texttt>i</texttt> = 0, 1,&#8230;,(<I>n</I> - 1)</SPAN>,
    * <SPAN CLASS="MATH"><texttt>j</texttt> = 0, 1,&#8230;,(<I>k</I> - 1)</SPAN>, the method
    * computes and returns the array
    * <SPAN CLASS="MATH">[<I>&#946;</I><SUB>0</SUB>, <I>&#946;</I><SUB>1</SUB>,&#8230;, <I>&#946;</I><SUB>k</SUB>]</SPAN>.
    * Restriction: <SPAN CLASS="MATH"><I>n</I> &gt; <I>k</I> + 1</SPAN>.
    *
    * @param X the regressor variables
    *
    *    @param Y the response
    *
    *    @return the regression coefficients
    *
    */
   public static double[] calcCoefficients0 (double[][] X, double[] Y) {
      if (X.length != Y.length)
         throw new IllegalArgumentException ("Lengths of X and Y are not equal");
      if (Y.length <= X[0].length + 1)
         throw new IllegalArgumentException ("Not enough points");

      final int n = Y.length;
      final int k = X[0].length;

      DoubleMatrix2D Xa = new DenseDoubleMatrix2D(n, k+1);
      DoubleMatrix2D Ya = new DenseDoubleMatrix2D(n, 1);

      for (int i = 0; i < n; i++) {
         Xa.setQuick (i, 0, 1.);
         for (int j = 1; j <= k; j++) {
            Xa.setQuick (i, j, X[i][j-1]);
         }
         Ya.setQuick (i, 0, Y[i]);
      }

      QRDecomposition qr = new QRDecomposition(Xa);
      DoubleMatrix2D Be = qr.solve(Ya);

      double[] Beta = new double[k+1];
      for (int j = 0; j <= k; j++)
         Beta[j] = Be.getQuick (j, 0);
      return Beta;
   }



   /**
    * Computes the regression coefficients using the
    * least squares method. This is a model for multiple linear regression.
    * There are <SPAN CLASS="MATH"><I>k</I></SPAN> regression coefficients <SPAN CLASS="MATH"><I>&#946;</I><SUB>j</SUB></SPAN>,
    * <SPAN CLASS="MATH"><I>j</I> = 0, 1,&#8230;,(<I>k</I> - 1)</SPAN> and
    * <SPAN CLASS="MATH"><I>k</I></SPAN> regressors variables <SPAN CLASS="MATH"><I>x</I><SUB>j</SUB></SPAN>. The model is
    *
    * <P></P>
    * <DIV ALIGN="CENTER" CLASS="mathdisplay">
    * <I>y</I> = &sum;<SUB>j=0</SUB><SUP>k-1</SUP><I>&#946;</I><SUB>j</SUB><I>x</I><SUB>j</SUB>.
    * </DIV><P></P>
    * There are <SPAN CLASS="MATH"><I>n</I></SPAN> data points <SPAN CLASS="MATH"><I>Y</I><SUB>i</SUB></SPAN>, <SPAN CLASS="MATH"><I>X</I><SUB>i, j</SUB></SPAN>,
    * <SPAN CLASS="MATH"><I>i</I> = 0, 1,&#8230;,(<I>n</I> - 1)</SPAN>, and each
    * <SPAN CLASS="MATH"><I>X</I><SUB>i</SUB></SPAN> is a <SPAN CLASS="MATH"><I>k</I></SPAN>-dimensional point.
    * Given the response <TT>Y[i]</TT> and the regressor variables <TT>X[i][j]</TT>,
    *
    * <SPAN CLASS="MATH"><texttt>i</texttt> = 0, 1,&#8230;,(<I>n</I> - 1)</SPAN>,
    * <SPAN CLASS="MATH"><texttt>j</texttt> = 0, 1,&#8230;,(<I>k</I> - 1)</SPAN>, the method
    * computes and returns the array
    * <SPAN CLASS="MATH">[<I>&#946;</I><SUB>0</SUB>, <I>&#946;</I><SUB>1</SUB>,&#8230;, <I>&#946;</I><SUB>k-1</SUB>]</SPAN>.
    * Restriction: <SPAN CLASS="MATH"><I>n</I> &gt; <I>k</I></SPAN>.
    *
    * @param X the regressor variables
    *
    *    @param Y the response
    *
    *    @return the regression coefficients
    *
    */
   public static double[] calcCoefficients (double[][] X, double[] Y) {
      if (X.length != Y.length)
         throw new IllegalArgumentException ("Lengths of X and Y are not equal");
      if (Y.length <= X[0].length + 1)
         throw new IllegalArgumentException ("Not enough points");

      final int n = Y.length;
      final int k = X[0].length;

      DoubleMatrix2D Xa = new DenseDoubleMatrix2D(n, k);
      DoubleMatrix2D Ya = new DenseDoubleMatrix2D(n, 1);

      for (int i = 0; i < n; i++) {
         for (int j = 0; j < k; j++) {
            Xa.setQuick (i, j, X[i][j]);
         }
         Ya.setQuick (i, 0, Y[i]);
      }

      QRDecomposition qr = new QRDecomposition(Xa);
      DoubleMatrix2D Be = qr.solve(Ya);

      double[] Beta = new double[k];
      for (int j = 0; j < k; j++)
         Beta[j] = Be.getQuick (j, 0);
      return Beta;
   }


}
